/* $Id: beacon.cpp 411 2010-05-06 19:29:33Z oh2gve $
 * 
 * Copyright 2009, 2010 Tapio Aaltonen
 *
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/

// Build config.
#ifdef HAVE_CONFIG_H
#	ifdef __WXMSW__
#		include "../windows/config.h"
#	else
#		include <config.h>
#	endif
#endif

// own stuff.
#include "beacon.h"
#include "aprspacket.h"
#include "connection.h"

// std stuff.
#include <vector>
using std::vector;
#include <string>
using std::string;
#include <map>
using std::multimap;
using std::make_pair;
#include <ctime>
#include <cmath>
#include <fstream>
using std::ifstream;
#include <sstream>
using std::ostringstream;
using std::istringstream;

// wx stuff.
#include <wx/string.h>
#include <wx/log.h>


Beacon::Beacon(std::vector<Connection*> const& conns, Connection* uplink, Cfg::Configuration const& cfg, DupeFilter& dupefilter, Packetlog& packetlog):
beacons_(), dupefilter_(dupefilter), packetlog_(packetlog), error_()
{
	// Beacon packet body is same for all connections, so we generate it first.

	// Check if we do tx or only rx.
	char overlay = 'R';
	if ( uplink->inputEnabled() )
	{
		overlay = 'I';
	}

	// Separate latitude into whole degrees and minutes.
	int lat_deg = fabs(cfg.location.lat);
	double lat_min = (fabs(cfg.location.lat) - lat_deg) * 60.0;

	// Separate longitude into whole degrees and minutes.
	int lon_deg = fabs(cfg.location.lon);
	double lon_min = (fabs(cfg.location.lon) - lon_deg) * 60.0;
	
	// Construct the packet.
	ostringstream info;
	info << "!";
	info << string(wxString::Format(wxT("%02d%02.2f"), lat_deg, lat_min).ToAscii());
	if ( cfg.location.lat < 0 )
	{
		info << "S";
	}
	else
	{
		info << "N";
	}
	info << overlay;
	info << string(wxString::Format(wxT("%03d%02.2f"), lon_deg, lon_min).ToAscii());
	if ( cfg.location.lon < 0 )
	{
		info << "W";
	}
	else
	{
		info << "E";
	}
	info << "&";
	info << cfg.comment.ToUTF8();

	// Create uplink beacon packet if necessary.
	if ( uplink->outputEnabled() )
	{
		// Create path.
		vector<string> netpath;
		netpath.push_back("TCPIP*");
		netpath.push_back("qAI");
		string tmp;
		tmp = cfg.callsign.ToAscii();
		netpath.push_back(tmp);
		
		// Create the packet.
		BeaconPacket bp;
		bp.type = 'b';
		bp.packet = new AprsPacket(tmp, OWN_TARGET, netpath, info.str(), false);
		bp.conn = uplink;
		bp.connNo = 0;
		bp.interval = cfg.beaconInterval;
		
		// Check that we got valid packet.
		if ( bp.packet->error(tmp) )
		{
			error_ = string("Failed to create beacon packet for uplink: ") + tmp + "(" + bp.packet->getTnc2Packet() + ")";
		}
		
		// Schedule beaconing to happen ASAP.
		beacons_.insert( make_pair(0, bp) );
	}
	
	// Create RF beacons as necessary.
	if ( error_.empty() )
	{
		for ( unsigned int c = 0; c < cfg.connections.size(); ++c )
		{
			// Check if this conn does TX.
			if ( cfg.connections.at(c).state != Cfg::READONLY )
			{
				// Create path.
				vector<string> rfpath;
				for ( unsigned int i = 0; i < cfg.connections.at(c).path.size(); i++ )
				{
					rfpath.push_back(string(cfg.connections.at(c).path.at(i).ToAscii()));
				}
				
				// Create the packet.
				BeaconPacket bp;
				bp.type = 'b';
				bp.packet = new AprsPacket(string(cfg.callsign.ToAscii()), OWN_TARGET, rfpath, info.str(), true);
				bp.conn = conns.at(c);
				bp.connNo = c+1;
				bp.interval = cfg.beaconInterval;
				
				// Check that we got valid packet.
				string tmp;
				if ( bp.packet->error(tmp) )
				{
					ostringstream errmsg;
					errmsg << "Failed to create beacon packet for connection " << c+1 << ": " << tmp;
					error_ = errmsg.str();
					break;
				}
				
				// Schedule beaconing to happen ASAP.
				beacons_.insert( make_pair(0, bp) );
			}
		}
	}
	
	// Then objects.
	if ( error_.empty() && !cfg.objectFile.empty() )
	{
		// Construct file name.
		string failnaem = "";
#ifndef __WXMSW__
		failnaem = string(SYSCONFDIR) + string("/");
#endif
		failnaem += string(cfg.objectFile.ToAscii());
		
		// Open file.
		wxLogMessage(wxT("Opening object config file ") + wxString::FromAscii(failnaem.c_str()));
		ifstream fail (failnaem.c_str());
		if ( !fail )
		{
			wxLogMessage(wxT("Failed to open!"));
			error_ = "Failed to open object config file.";
		}
		else
		{
			// Read all lines.
			string line;
			unsigned int lineno = 0;
			while ( getline(fail, line) )
			{
				lineno++;
				
				// Check for too short or comment line.
				if ( line.length() < 31 || line.at(0) == '#' )
				{
					continue;
				}
				
				// Get offset.
				string offset = "";
				while ( !line.empty() && isdigit(line.at(0)) )
				{
					offset += line.at(0);
					line.erase(0,1);
				}
				
				// Skip space.
				if ( line.empty() || line.at(0) != ' ' )
				{
					error_ = "Line \"" + int2string(lineno) + "\" of object file is invalid.";
					break;
				}
				line.erase(0,1);
				
				// Get interval.
				string interval = "";
				while ( !line.empty() && isdigit(line.at(0)) )
				{
					interval += line.at(0);
					line.erase(0,1);
				}
				
				// Skip space.
				if ( line.empty() || line.at(0) != ' ' )
				{
					error_ = "Line \"" + int2string(lineno) + "\" of object file is invalid.";
					break;
				}
				line.erase(0,1);
				
				// Get path.
				vector<string> opath;
				string pElem = "";
				while ( !line.empty() && line.at(0) != ' ' )
				{
					if ( line.at(0) == ',' )
					{
						opath.push_back(pElem);
						pElem.clear();
					}
					else
					{
						pElem += line.at(0);
					}
					line.erase(0,1);
				}
				if ( !line.empty() && !pElem.empty() && pElem != "nopath" )
				{
					opath.push_back(pElem);
				}
				
				// Skip space.
				if ( line.empty() || line.at(0) != ' ' )
				{
					error_ = "Line \"" + int2string(lineno) + "\" of object file is invalid.";
					break;
				}
				line.erase(0,1);
				
				// Rest is the object.
				
				// Check that packet compiles.
				AprsPacket* tmpPacket = new AprsPacket(string(cfg.callsign.ToAscii()), OWN_TARGET, opath, line, true);
				string tmp;
				if ( tmpPacket->error(tmp) )
				{
					ostringstream errmsg;
					errmsg << "Failed to create carrier packet for object file line " << lineno << ": " << tmp;
					error_ = errmsg.str();
					delete(tmpPacket);
					break;
				}
				wxLogMessage(wxT("Scheduling transmission of object ") + wxString::FromAscii(tmpPacket->getLabel().c_str()) );
				delete(tmpPacket);
				
				// Configure transmishun for uplink.
				if ( uplink->outputEnabled() )
				{
					// First we need path.
					vector<string> opnp;
					opnp.push_back(string("TCPIP*"));
					
					// Then da packet.
					BeaconPacket op;
					op.type = 'o';
					op.packet = new AprsPacket(string(cfg.callsign.ToAscii()), OWN_TARGET, opnp, line, true);
					op.interval = 60*string2int(interval);
					op.conn = uplink;
					op.connNo = 0;
					beacons_.insert( make_pair(time(NULL)+60*string2int(offset), op) );
				}
				
				// And for each outputting con.
				for ( unsigned int i = 0; i < conns.size(); ++i )
				{
					if ( conns.at(i)->outputEnabled() )
					{
						BeaconPacket op;
						op.type = 'o';
						op.packet = new AprsPacket(string(cfg.callsign.ToAscii()), OWN_TARGET, opath, line, true);
						op.interval = 60*string2int(interval);
						op.conn = conns.at(i);
						op.connNo = i+1;
						beacons_.insert( make_pair(time(NULL)+60*string2int(offset), op) );
					}
				}
			} // while lines
		} // file opened
		fail.close();
	} // no error
}



Beacon::~Beacon()
{
	for ( multimap<time_t, BeaconPacket>::iterator i = beacons_.begin();
			i != beacons_.end(); ++i )
	{
		delete(i->second.packet);
	}
}



bool Beacon::operator!() const
{
	return !error_.empty();
}


string Beacon::error() const
{
	return error_;
}



void Beacon::kick()
{
	// Check if next beaconing is at hand.
	if ( !beacons_.empty() && time(NULL) >= beacons_.begin()->first )
	{
		// Put the packet in dupe control.
		dupefilter_.isDupe(beacons_.begin()->second.packet);
		
		// Create a copy of the packet to be transmitted.
		AprsPacket* output = new AprsPacket(*beacons_.begin()->second.packet);
		
		// Transmit the packet.
		beacons_.begin()->second.conn->write(output);
		
		// Create a copy of the packet for packet log.
		Packet* logPkt = new Packet();
		logPkt->timestamp = wxDateTime::Now();
		logPkt->input = '_';
		logPkt->aprsPacket = new AprsPacket(*beacons_.begin()->second.packet);
		logPkt->thirdParty = NULL;
		logPkt->output = beacons_.begin()->second.connNo + 48;
		if ( beacons_.begin()->second.connNo == 0 )
		{
			logPkt->output = 'u';
		}
		logPkt->flag = beacons_.begin()->second.type;
		
		// Put in log.
		packetlog_.write(logPkt);
		
		// Re-schedule the beacon.
		beacons_.insert( make_pair(time(NULL)+beacons_.begin()->second.interval, beacons_.begin()->second) );
		
		// Unschedule this beacon.
		beacons_.erase(beacons_.begin());
	}
}


string Beacon::int2string(unsigned int const& i) const
{
	ostringstream s;
	s << i;
	return s.str();
}

unsigned int Beacon::string2int(string const& i) const
{
	istringstream s (i);
	unsigned int retval;
	s >> retval;
	return retval;
}
