/* $Id: filter.cpp 408 2010-05-05 12:09:42Z oh2gve $
 *
 * Copyright 2007, 2008, 2009, 2010 Tapio Aaltonen 
 * 
 * 
 * This file is part of aprsg.
 *
 * aprsg is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * aprsg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with aprsg.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * 
 * APRS is a registered trademark of APRS Software and Bob Bruninga, WB4APR.
*/


// Own stuff.
#include "filter.h"

// std stuff.
#include <vector>
using std::vector;
#include <cmath>

// wx stuff.
#include <wx/log.h>
#include <wx/string.h>
#include <wx/regex.h>


// Conversion constants
#define DEG2RAD(x) (x/360*2*3.14159265)



namespace Cfg
{
	Filter::Filter():
	areas_(), ranges_(), buddyRegs_(), buddyPatterns_()
	{
	}
	
	Filter::Filter(Filter const& old):
	areas_(old.areas_), ranges_(old.ranges_), buddyRegs_(), buddyPatterns_(old.buddyPatterns_)
	{
		for ( unsigned int i = 0; i < old.buddyRegs_.size(); i++ )
		{
			wxRegEx* tmp = new wxRegEx(old.buddyPatterns_.at(i));
			buddyRegs_.push_back(tmp);
		}
	}
	
	Filter::~Filter()
	{
		for ( unsigned int i = 0; i < buddyRegs_.size(); i++ )
		{
			delete buddyRegs_.at(i);
		}
	}
	
	
	void Filter::addArea(Area const& a)
	{
		areas_.push_back(a);
	}
	
	void Filter::addRange(Range const& r)
	{
		ranges_.push_back(r);
	}
		
	void Filter::addBuddyPattern(wxString const& p)
	{
		buddyPatterns_.push_back(p);
		buddyRegs_.push_back(new wxRegEx(p));
	}
	
	
	
	bool Filter::hasBuddyPatterns() const
	{
		return !buddyRegs_.empty();
	}
	
	unsigned int Filter::getPartCount() const
	{
		return areas_.size() + buddyRegs_.size() + ranges_.size();
	}

	
	bool Filter::accepts(std::string const& call, double const& lon, double const& lat) const
	{
		// Convert callsign into wxString for matching.
		wxString callsign = wxString::FromAscii(call.c_str());
		
		// Check range.
		if ( !ranges_.empty() )
		{
			bool match = false;
			for ( unsigned int i = 0; i < ranges_.size(); ++i )
			{
				// If packet is in range, it's good.
				double range = distance(lon, lat, ranges_.at(i).lon, ranges_.at(i).lat);
				if ( range <= ranges_.at(i).dist )
				{
					match = true;
					break;
				}
			}
			if ( !match )
			{
				return false;
			}
		}
		
		// Check callsign patterns for matches.
		if ( !buddyRegs_.empty() )
		{
			bool match = false;
			for ( unsigned int i = 0; i < buddyRegs_.size(); i++ )
			{
				if ( buddyRegs_.at(i)->Matches(callsign) )
				{
					match = true;
					break;
				}
			}
			if ( !match )
			{
				return false;
			}
		}
		
		// Check area.
		if ( !areas_.empty() )
		{
			bool match = false;
			for ( unsigned int i = 0; i < areas_.size(); ++i )
			{
				// If packet is in this area, it's going to pass.
				if ( lat <= areas_.at(i).latN && lat >= areas_.at(i).latS &&
					  lon >= areas_.at(i).lonW && lon <= areas_.at(i).lonE )
				{
					match = true;
					break;
				}
			}
			if ( !match )
			{
				return false;
			}
		}

		return true;
	}
	
	
	bool Filter::accepts(std::string const& call) const
	{
		// Convert callsign into wxString for matching.
		wxString callsign = wxString::FromAscii(call.c_str());
		
		// If range or area parts are available, we fail straight away.
		if ( !ranges_.empty() || !areas_.empty() )
		{
			return false;
		}
		
		// Check callsign patterns for matches.
		if ( !buddyRegs_.empty() )
		{
			bool match = false;
			for ( unsigned int i = 0; i < buddyRegs_.size(); i++ )
			{
				if ( buddyRegs_.at(i)->Matches(callsign) )
				{
					match = true;
					break;
				}
			}
			if ( !match )
			{
				return false;
			}
		}

		return true;
	}
	
	
	// PRIVATE
	 
	 
	double Filter::distance(double lon0, double lat0, double lon1, double lat1) const
	{
		// Convert input values into radians.
		lon0 = DEG2RAD(lon0);
		lat0 = DEG2RAD(lat0);
		lon1 = DEG2RAD(lon1);
		lat1 = DEG2RAD(lat1);
	 
		// Use the haversine formula for distance calculation
		// http://mathforum.org/library/drmath/view/51879.html
		double dlon = lon1 - lon0;
		double dlat = lat1 - lat0;
		double a = pow(sin(dlat/2),2) + cos(lat0) * cos(lat1) * pow(sin(dlon/2), 2);
		double c = 2 * atan2(sqrt(a), sqrt(1-a));
		
		return c * 6366.71;
	}
		
} // namespace Cfg
